# -*- coding: utf-8 -*-
"""Rubix :class:`Cube_Game` class Module

Module Description
==================

Collection of methods that define the main Rubix :class:`Cube Game <Cube_Game>`
class and how it manipulates a :class:`Cube <rubix_cube.cube.Cube>`
object, prints to console, logs a history of both moves and re-orientations, 
scrambles the state of the cube, times game-play, and performs file-system
loading and saving of games.


Module Contents
===============

    *   :class:`Cube_Game` details code for manipulating :class:`Rubix 
        Cube(s) <rubix_cube.cube.Cube>`.

        .. todo::

           *    Implement file I/O operations (loading and saving games).
           *    Implement scramble cube functions.
           *    Implement color changing functions.
           *    Implement game clock functions.


.. moduleauthor:: David Grethlein <djg329@drexel.edu>

"""

from pathlib import Path 
from random import randint , randrange

import os
import sys
import json

import warnings

from typing import List, Tuple, Dict

import numpy as np

from .cube import Cube


class Cube_Game(object):
    """Class in charge of directly interacting with and logging changes to
    an instance of the :class:`Cube` class.

    Attributes:
        EVENT_TYPES (List[str]): Description
        __game_cube (Cube): :class:`Rubix Cube <Cube>` that is manipulated
            throughout game-play.
        __game_log (dict): Historical record of all moves, rotations, and other
            game events that manipulate the :attr:`game_cube`.
        __game_name (str): Name of the game.
        __verbose (bool): [DEBUG]-style console output. Default value is 
            ``False``.
    
    """

    #==========================================================================
    #       CONSTANTS FOR CUBE GAME(s)
    #==========================================================================
    EVENT_TYPES = ['<<__NEW_GAME__>>',
                   'F',
                   'Fi',
                   'B',
                   'Bi',
                   'L',
                   'Li',
                   'R',
                   'Ri',
                   'U',
                   'Ui',
                   'D',
                   'Di',
                   'M',
                   'Mi',
                   'E',
                   'Ei',
                   'S',
                   'Si',
                   'X',
                   'Xi',
                   'Y',
                   'Yi',
                   'Z',
                   'Zi',
                   '<<__START_SCRAMBLE__>>',
                   '<<__END_SCRAMBLE__>>',
                   '<<__COLOR_CHANGE__>>',
                   '<<__SAVE_GAME__>>',
                   '<<__LOAD_GAME__>>',
                   '<<__PAUSE_GAME__>>',
                   '<<__RESUME_GAME__>>',
                   '<<__SOLVE_CUBE__>>',
                   '<<__QUIT_GAME__>>']


    CUBE_FUNCS = {'U'  : Cube.move_up,
                  'Ui' : Cube.move_up_inverse,
                  'D'  : Cube.move_down,
                  'Di' : Cube.move_down_inverse,
                  'L'  : Cube.move_left,
                  'Li' : Cube.move_left_inverse,
                  'R'  : Cube.move_right,
                  'Ri' : Cube.move_right_inverse,
                  'F'  : Cube.move_front,
                  'Fi' : Cube.move_front_inverse,
                  'B'  : Cube.move_back,
                  'Bi' : Cube.move_back_inverse,
                  'M'  : Cube.move_middle,
                  'Mi' : Cube.move_middle_inverse,
                  'E'  : Cube.move_equator,
                  'Ei' : Cube.move_equator_inverse,
                  'S'  : Cube.move_standing,
                  'Si' : Cube.move_standing_inverse,
                  'X'  : Cube.rotate_pitch,
                  'Xi' : Cube.rotate_pitch_inverse,
                  'Y'  : Cube.rotate_yaw,
                  'Yi' : Cube.rotate_yaw_inverse,
                  'Z'  : Cube.rotate_roll,
                  'Zi' : Cube.rotate_roll_inverse}

    INVERSE_FUNCS = {'U'  : 'Ui',
                     'Ui' : 'U',
                     'D'  : 'Di',
                     'Di' : 'D',
                     'L'  : 'Li',
                     'Li' : 'L',
                     'R'  : 'Ri',
                     'Ri' : 'R',
                     'F'  : 'Fi',
                     'Fi' : 'F',
                     'B'  : 'Bi',
                     'Bi' : 'B',
                     'M'  : 'Mi',
                     'Mi' : 'M',
                     'E'  : 'Ei',
                     'Ei' : 'E',
                     'S'  : 'Si',
                     'Si' : 'S',
                     'X'  : 'Xi',
                     'Xi' : 'X',
                     'Y'  : 'Yi',
                     'Yi' : 'Y',
                     'Z'  : 'Zi',
                     'Zi' : 'Z'}

    #==========================================================================
    #       CLASS CONSTRUCTOR
    #==========================================================================
    def __init__(self,
                 cube : Cube = None,
                 game_name : str = None,
                 game_log : Dict = None,
                 scramble : bool = False,
                 verbose : bool = False):
        """:class:`Cube_Game` class constructor
        
        Args:
            cube (Cube, optional): :class:`Cube` that will be directly
                manipulated throughout gameplay.
            game_name (str, optional): Name of the current game being played.
            game_log (Dict, optional): Dictionary that contains a history of
                moves and other game events.
            scramble (bool, optional): Whether or not the game should scramble 
                the :attr:`__game_cube` upon initialization. Default value is
                ``False``.
            verbose (bool, optional): [DEBUG]-style console output. Default
                value is ``False``.
        
        """

        # Sets Up Default Game
        self.game_name = 'Untitled_Cube_Game'
        self.game_cube = Cube()
        self.game_log = {'events' : [{'type' : '<<__NEW_GAME__>>',
                              'name' : self.game_name}]}

        # Attempts to reset property values with argument values.
        self.game_cube = cube
        self.game_name = game_name
        self.game_log = game_log
        self.verbose = verbose

        # Initializes a default cube
        if self.game_cube == Cube():
            
            if self.verbose:
                print(f"\n[DEBUG]\tNew DEFAULT Cube created for game : '{self.game_name}'\n")

        # Initializes a default game log
        if self.game_log != game_log:

            if self.verbose:
                print(f"\n[DEBUG]\tNew DEFAULT ``game_log`` created for game : '{self.game_name}'\n")
        
        else:
            self.game_log['events'].append({'type' : '<<__NEW_GAME__>>',
                                            'name' : self.game_name})

            if self.verbose:
                print(f"\n[DEBUG]\tNew game created with name : '{self.game_name}'\n")


    #==========================================================================
    #       PROPERTY INTERFACE(s)
    #==========================================================================
    @property
    def game_cube(self) -> Cube:
        """:class:`Rubix Cube <Cube>` object being manipulated in game.
        """
        return self.__game_cube


    @game_cube.setter
    def game_cube(self , cube : Cube):
        if isinstance(cube, Cube)\
        and cube.is_well_formed():

            self.__game_cube = cube


    @property
    def game_name(self) -> str:
        """Name of the Game
        """
        return self.__game_name
    

    @game_name.setter
    def game_name(self, name : str):
        if isinstance(name, str)\
        and len(name) > 0:

            self.__game_name = name


    @property
    def game_log(self) -> Dict:
        """JSON-style :class:`dict` recording all actions done to
        the :attr:`game_cube` stored under the ``events`` key which is a list
        of :class`dict` objects each of which has a ``type`` key with a value
        found in :attr:`EVENT_TYPES`.

        .. code-block::
           :name: game_log_EVENT_TYPES
           :linenos:
           :caption: Potential values of ``type`` in :attr:`game_log`.

           EVENT_TYPES = ['<<__NEW_GAME__>>',
                          'F',
                          'Fi',
                          'B',
                          'Bi',
                          'L',
                          'Li',
                          'R',
                          'Ri',
                          'U',
                          'Ui',
                          'D',
                          'Di',
                          'M',
                          'Mi',
                          'E',
                          'Ei',
                          'S',
                          'Si',
                          'X',
                          'Xi',
                          'Y',
                          'Yi',
                          'Z',
                          'Zi',
                          '<<__START_SCRAMBLE__>>',
                          '<<__END_SCRAMBLE__>>',
                          '<<__COLOR_CHANGE__>>',
                          '<<__SAVE_GAME__>>',
                          '<<__LOAD_GAME__>>',
                          '<<__PAUSE_GAME__>>',
                          '<<__RESUME_GAME__>>',
                          '<<__SOLVE_CUBE__>>',
                          '<<__QUIT_GAME__>>']
        
        .. code-block::
           :name: game_log
           :linenos:
           :caption: Required ``game_log`` dictionary keys.

           game_log = {'events' : [{'type' : '<<__NEW_GAME__>>',
                                    'name' : '...'},
                                   {'type' : ...},
                                   ...,
                                   {'type' : ...}]}
        """
        return self.__game_log


    @game_log.setter
    def game_log(self, game_log : Dict):
        if isinstance(game_log, dict)\
        and 'events' in game_log\
        and isinstance(game_log['events'], list)\
        and len(game_log['events']) > 0\
        and all([isinstance(event, dict) for event in game_log['events']]):

            # Has to ensure that all event types are valid
            valid_log = all(['type' in event\
                             and event['type'] in Cube_Game.EVENT_TYPES\
                                for event in game_log['events']])

            if valid_log:
                self.__game_log = game_log


    @property
    def verbose(self) -> bool:
        """[DEBUG]-style console output. Default value is ``False``.
        """
        return self.__verbose


    @verbose.setter
    def verbose(self, verbose : bool):
        if isinstance(verbose, bool):
            self.__verbose = verbose
        else:
            self.__verbose = False 


    #==========================================================================
    #       GAME-PLAY METHOD(s)
    #==========================================================================
    def manipulate_cube(self, cube_func : str):
        """Function that interfaces the :class:`Cube_Game` class with the
        :class:`Cube` class to turn the layers or rotate the orientation.

        Args:
            cube_func (str): Look-up key to recover the proper :class:`Cube` 
                method to call from :attr:`CUBE_FUNCS` class attribute to call
                a move using the :attr:`game_cube`.

        .. code-block::
           :name: move_cube_CUBE_FUNCS
           :linenos:
           :caption: Parameter ``cube_func`` will determine which 
                :attr:`game_cube` move function is called. If not found,
                nothing happens.

           CUBE_FUNCS = {'U'  : Cube.move_up,
                         'Ui' : Cube.move_up_inverse,
                         'D'  : Cube.move_down,
                         'Di' : Cube.move_down_inverse,
                         'L'  : Cube.move_left,
                         'Li' : Cube.move_left_inverse,
                         'R'  : Cube.move_right,
                         'Ri' : Cube.move_right_inverse,
                         'F'  : Cube.move_front,
                         'Fi' : Cube.move_front_inverse,
                         'B'  : Cube.move_back,
                         'Bi' : Cube.move_back_inverse,
                         'M'  : Cube.move_middle,
                         'Mi' : Cube.move_middle_inverse,
                         'E'  : Cube.move_equator,
                         'Ei' : Cube.move_equator_inverse,
                         'S'  : Cube.move_standing,
                         'Si' : Cube.move_standing_inverse,
                         'X'  : Cube.rotate_pitch,
                         'Xi' : Cube.rotate_pitch_inverse,
                         'Y'  : Cube.rotate_yaw,
                         'Yi' : Cube.rotate_yaw_inverse,
                         'Z'  : Cube.rotate_roll,
                         'Zi' : Cube.rotate_roll_inverse}
        """

        if cube_func in Cube_Game.CUBE_FUNCS\
        and self.game_cube.is_well_formed():

            # Performs the desired cube move on the game cube
            Cube_Game.CUBE_FUNCS[cube_func](self.game_cube)

            if self.verbose:
                print(f"[DEBUG]\tCalling game cube function: '{cube_func}'")

            self.game_log['events'].append({'type' : cube_func})


    
    def get_scramble_sequence(n_steps : int = 50,
                              cube_funcs : List[str] = None) -> List[str]:
        """Compiles a sequence of moves for scrambling the :attr:`game_cube`
        attribute for applying a sequence of ``n_steps`` semi-randomly selected
        cube manipulations using a defined a provided sub-set of cube functions
        ``cube_funcs``.

        Args:
            n_steps (int, optional): The number of :class:`Rubix Cube <Cube>` 
                manipulations to be applied to the :attr:`game_cube`.

                Note:
                    Valid range of values, 0 < ``n_steps`` <= 500. Won't 
                    call any :class:`Cube` function(s) if not in range.

            cube_funcs (List[str], optional): Sub-list of :attr:`CUBE_FUNCS`
                that defines the options for manipulating the cube. Default
                value is ``None`` which allows all functions in 
                :attr:`CUBE_FUNCS` to be selected.
        
        Returns:
            List[str]: **sequence** - A list of cube manipulation function 
            str(s) for use with :func:`manipulate_cube` function.

        Note:
            Won't perform the following sequences:

            *   ``<<ACTION>>`` , ``<<ACTION_INVERSE>>``
            *   ``<<ACTION>>``, ``<<ACTION>>``, ``<<ACTION>>``, ``<<ACTION>>``
        
            .. code-block::
               :name: scramble_cube_INVERSE_FUNCS
               :linenos:
               :caption: :class:`Cube_Game` static :attr:`INVERSE_FUNCS` attribute
                    for looking up the inverse function for each potential cube
                    manipulation.
            
               INVERSE_FUNCS = {'U'  : 'Ui',
                                'Ui' : 'U',
                                'D'  : 'Di',
                                'Di' : 'D',
                                'L'  : 'Li',
                                'Li' : 'L',
                                'R'  : 'Ri',
                                'Ri' : 'R',
                                'F'  : 'Fi',
                                'Fi' : 'F',
                                'B'  : 'Bi',
                                'Bi' : 'B',
                                'M'  : 'Mi',
                                'Mi' : 'M',
                                'E'  : 'Ei',
                                'Ei' : 'E',
                                'S'  : 'Si',
                                'Si' : 'S',
                                'X'  : 'Xi',
                                'Xi' : 'X',
                                'Y'  : 'Yi',
                                'Yi' : 'Y',
                                'Z'  : 'Zi',
                                'Zi' : 'Z'}

        """
    
        # Default length for sequence of random moves to be generated
        if not isinstance(n_steps, int)\
        or not (n_steps > 0 and n_steps <= 500):
            n_steps = 0

        # Default list of cube functions is all of them
        if not isinstance(cube_funcs , list)\
        or not all([func in Cube_Game.CUBE_FUNCS for func in cube_funcs]):
            cube_funcs = list(Cube_Game.CUBE_FUNCS.keys())

        sequence = []

        while len(sequence) < n_steps:

            # Generates a random choice from the current options
            # of potential cube-functions to call
            choice_func = cube_funcs[randrange(len(cube_funcs))]

            if len(sequence) == 0:
                sequence.append(choice_func)
            else:
                last_func = sequence[-1]

                # No <<FUNCTION>> , <<INVERSE_FUNCTION>> sub-sequences
                if last_func == Cube_Game.INVERSE_FUNCS[choice_func]:
                    continue

                if len(sequence) > 2:

                    last_three = sequence[-3:]

                    # No 4 of the same function in a row
                    if all([func == choice_func for func in last_three]):
                        continue

            sequence.append(choice_func)

        return sequence








# -*- coding: utf-8 -*-
"""2-D :class:`Rubix Cube <Cube>` plot rendering Module.

Module Description
==================

    *   :func:`plot_cube_2D` method for plotting the :class:`Rubix Cube <Cube>`
        in a :class:`matplotlib Figure <matplotlib.pyplot.Figure>`.

        .. figure:: ./../../misc/plot_cube_2D.png
           :name: plot_cube_2d_img
           :align: center
           :width: 75%
           
           All 6 faces of the :class:`Rubix Cube <Cube>` are visible in this 2D
           rendering.

Module Contents
===============

.. moduleauthor:: David Grethlein <djg329@drexel.edu>

"""

import os
import sys
import json

import warnings

from typing import List, Tuple, Dict

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches

from .cube import Cube
from .cube_game import Cube_Game


FACE_COORDS = {'FRONT_FACE' : (-150,50),
               'LEFT_FACE'  : (-450,50),
               'RIGHT_FACE' : (150,50),
               'BACK_FACE'  : (450,50),
               'UP_FACE'    : (-150,350),
               'DOWN_FACE'  : (-150,-250)}


def plot_cube_2D(ax : plt.Axes , cube : Cube):
    """Plots the :class:`Rubix Cube <Cube>` on a :class:`matplotlib.Figure`
    in a flattened 2-D representation.

    Args:
        ax (plt.Axes): The axes on which the plotting will be done.
        cube (Cube): :class:`Rubix Cube <Cube>` that will be plotted in a
            flattened 2-D representation of all 6 sides.

    Returns:
        None.

    """

    # Ensures the arguments are well-formed.
    if cube.is_well_formed()\
    and isinstance(ax , plt.Axes):

        for face_name in cube.faces:

            # Grabs the local coordinates for each face
            face_local_coords = FACE_COORDS[face_name]

            # Grabs the face values for filling in the plot for each tile
            # in the 2-D representation.
            face_values = cube.faces[face_name]

            #print(f"{face_name} : ")

            for row_idx , row in enumerate(face_values):

                for col_idx , color_val in enumerate(row):

                    tile_coords = (face_local_coords[0] + 100*col_idx,
                                   face_local_coords[1] - 100*row_idx)

                    #print(f"\t[{row_idx},{col_idx}] : {tile_coords}")

                    rect = patches.Rectangle(xy=tile_coords,
                                             width=100,
                                             height=100,
                                             facecolor=color_val,
                                             lw=2,
                                             edgecolor='black')

                    ax.add_patch(rect)






#!/usr/bin/env python

from distutils.core import setup

RELEASE_VERSION='0.1.6'

setup(name = 'rubix_cube',
      packages = ['rubix_cube'],   
      version = RELEASE_VERSION,    
      license='MIT',  
      description = 'Python Rubix Cube GUI package',
      author = 'David Grethlein',
      author_email = '12dgrethlein@gmail.com',
      url = 'https://github.com/dgrethlein/RubixCube' , 
      download_url = f'https://github.com/dgrethlein/RubixCube/archive/v{RELEASE_VERSION}.tar.gz',
      keywords = ['Python', 'GUI', 'Rubix' , 'A*'],
      install_requires = ['numpy' , 'pandas' , 'matplotlib'],
      classifiers = ['Development Status :: 3 - Alpha',
        'Intended Audience :: Developers', 
        'Intended Audience :: End Users/Desktop', 
        'Topic :: Games/Entertainment :: Arcade',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
      ],
)


# -*- coding: utf-8 -*-
"""Rubix :class:`Cube` class data-structure Module

Module Description
==================

Collection of methods that define the main Rubix :class:`Cube` class data
structure and how it is interacted with by other modules.

Note:
    Using `Western Color Scheme 
    <https://ruwix.com/the-rubiks-cube/japanese-western-color-schemes/>`_ as
    default Rubix Cube coloring scheme.

Module Contents
===============

    *   :class:`Rubix Cube <Cube>` class that is capable of being parameterized
        with a custom set of 6 unique colors (`Default Color Scheme 
        <https://www.schemecolor.com/rubik-cube-colors.php>`_) and can invoke
        the following moves.

        .. figure:: ./../../misc/cube_moves.png
           :name: cube_moves
           :align: center
           :scale: 75%

           6 cube face rotations both clock-wise and 
           counter-clockwise (inverse) are considered to be the standard 
           move-set. (`How to Solve 
           <https://www.rubiks.com/en-us/blog/how-to-solve-the-rubiks-cube-stage-1>`_
           )

        .. todo::
           
           *    Need to finish implementing the ``get_num_solved_rings``.

        .. figure:: ./../../misc/flattened_cube.png
           :name: flattened_cube
           :align: center
           :scale: 45%

           At this view of the cube, each face is a 3x3 array indexed with 
           [0,0] in the top-left and [2,2] in the bottom right. (`Flattened
           <https://rantonse.no/en/blog/2016-05-12>`_)

.. moduleauthor:: David Grethlein <djg329@drexel.edu>

"""

import os
import sys
import json

import warnings

from typing import List, Tuple, Dict

import numpy as np
from matplotlib.colors import is_color_like


class Cube(object):
    """Data structure for representing a 3x3x3 rubix-cube.

    Attributes:
        __colors (Dict[str,str]): Dictionary of HEX colors that define the 
            rendering of the :class:`Cube`'s tile coloring.
        __faces (Dict[str,np.ndarray]): Dictionary of 
            :class:`numpy arrays <numpy.ndarray>` that define the rendering of
            the :class:`Cube`'s tile configuration.

    """

    #==========================================================================
    #       CLASS CONSTRUCTOR
    #==========================================================================
    def __init__(self,
                 colors : Dict[str,str] = None,
                 faces : Dict[str,np.array] = None):
        """:class:`Cube` class constructor.

        Args:
            colors (Dict[str,str], optional): Dictionary of color HEX strings.
                Default value is ``None`` which will create a cube with default
                colors :attr:`DEFAULT_FACE_COLORS`. 

                .. code-block:: 
                   :name: init_colors_keys
                   :linenos:
                   :caption: Required ``colors`` dictionary keys.

                   colors = {'UP_COLOR' : ...,
                             'DOWN_COLOR' : ...,
                             'FRONT_COLOR' : ...,
                             'BACK_COLOR' : ...,
                             'LEFT_COLOR' : ...,
                             'RIGHT_COLOR' : ...
                            }

                All colors passed as values must return ``True`` when examined
                by :func:`matplotlib.colors.is_color_like`.

            faces (Dict[str,np.array], optional): Dictionary of face names to
                3x3 arrays of the the tile face values. Default value is 
                ``None`` which will create a solved cube with default colors.

                .. code-block::
                   :name: init_faces_keys 
                   :linenos:
                   :caption: Required ``faces`` dictionary keys.

                   faces = {'UP_FACE' : ...,
                            'DOWN_FACE' : ...,
                            'FRONT_FACE' : ...,
                            'BACK_FACE' : ...,
                            'LEFT_FACE' : ...,
                            'RIGHT_FACE' : ...
                           }

                All faces passed as value must be 3x3 
                :class:`numpy arrays <numpy.ndarray>` with each element 
                returning ``True`` when examined by 
                :func:`matplotlib.colors.is_color_like`.

        """

        # Sets private attributes via properties
        self.colors = Cube.DEFAULT_FACE_COLORS
        self.faces = Cube.DEFAULT_FACES 
        
        if isinstance(colors, dict):
            self.colors = colors

            if isinstance(faces, dict):
                self.faces = faces          


    #==========================================================================
    #       PROPERTY INTERFACE(s)
    #==========================================================================
    @property
    def colors(self):
        """Can only be set to be a dictionary with 6 unique color string values
        that all return ``True`` when examined by 
        :func:`matplotlib.colors.is_color_like`. 

        .. code-block::
           :name: colors_keys
           :linenos:
           :caption: Required ``colors`` dictionary keys.
           
           colors = {'UP_COLOR' : ...,
                     'DOWN_COLOR' : ...,
                     'FRONT_COLOR' : ...,
                     'BACK_COLOR' : ...,
                     'LEFT_COLOR' : ...,
                     'RIGHT_COLOR' : ...
                    }
        """
        return self.__colors

    
    @colors.setter
    def colors(self, colors : Dict[str,str]):

        required_keys = ['UP_COLOR',
                         'DOWN_COLOR',
                         'FRONT_COLOR',
                         'BACK_COLOR',
                         'LEFT_COLOR',
                         'RIGHT_COLOR']

        if isinstance(colors, dict)\
        and all([key in colors for key in required_keys]):

            set_colors = np.array([colors[key] for key in required_keys])

            if all([is_color_like(color) for color in set_colors]):

                self.__colors = dict(zip(required_keys , set_colors))


    @property
    def faces(self) -> Dict[str,np.ndarray]:
        """Can only be set to be a dictionary of 6 strings mapped to the faces
        of a Rubix Cube. Each value must be a 3x3 
        :class:`numpy array <numpy.ndarray>` of values all of which are valid
        colors that can be found within the :attr:`colors` attribute.
        
        .. code-block::
           :name: faces_keys
           :linenos:
           :caption: Required ``faces`` dictionary keys.
           
           faces = {'UP_FACE' : ...,
                    'DOWN_FACE' : ...,
                    'FRONT_FACE' : ...,
                    'BACK_FACE' : ...,
                    'LEFT_FACE' : ...,
                    'RIGHT_FACE' : ...
                   }
        """
        return self.__faces


    @faces.setter
    def faces(self, faces : Dict[str,np.ndarray]):

        required_keys = ['UP_FACE',
                         'DOWN_FACE',
                         'FRONT_FACE',
                         'BACK_FACE',
                         'LEFT_FACE',
                         'RIGHT_FACE']
        
        if isinstance(faces, dict)\
        and all([key in faces for key in required_keys]):

            set_faces = np.array([faces[key] for key in required_keys])

            if all([self.is_valid_face(face) for face in set_faces]):

                self.__faces = dict(zip(required_keys , set_faces))

    #==========================================================================
    #       QUALITY ASSURANCE METHOD(s)
    #========================================================================== 
    def is_well_formed(self) -> bool:
        """Quality control method to ensure class has been properly
        initialized by examining all :attr:`faces` via the quality control
        method :func:`is_valid_face`.
        
        Returns:
            ``True`` if all faces are 3 x 3 arrays of valid colors
            as defined by :func:`matplotlib.colors.is_color_like`,
            ``False`` otherwise.

        """
        required_keys = ['UP_FACE',
                         'DOWN_FACE',
                         'FRONT_FACE',
                         'BACK_FACE',
                         'LEFT_FACE',
                         'RIGHT_FACE']

        try:
            return all([self.is_valid_face(self.faces[face]) 
                        for face in required_keys])

        except KeyError:
            return False


    def is_valid_face(self, face : np.ndarray) -> bool:
        """Checks if the provided array could be a valid face on the 
        currently initialized :class:`Cube`.
            
        Args:
            face (np.ndarray): Array to be tested for being valid in the
                context of the current :class:`Cube`.

        Returns:
            ``True`` if faces is 3 x 3 array of valid colors as defined by
            current instance's :attr:`colors` attribute, ``False`` otherwise.
        """
        if isinstance(face, np.ndarray)\
        and face.shape == (3,3)\
        and all([all([val in self.colors.values() 
                        for val in row])
                            for row in face]):

            return True

        else:

            return False


    def is_solved_face(self, face: np.ndarray) -> bool:
        """Checks if the provided array could be a valid face on the 
        currently initialized :class:`Cube`.

        Args:
            face (np.ndarray): Array to be tested for being solved in the
                context of the current :class:`Cube`.

        Returns:
            ``True`` if faces is solved 3 x 3 array of valid colors as defined
            by current instance's :attr:`colors` attribute, ``False``
            otherwise.

        .. code-block::
           :name: is_solved_face
           :linenos:
           :caption: A solved ``face`` returns ``True`` when examined by 
                :func:`is_valid_face` and only contains 1 unique value.

            return len(np.unique(face) == 1)
        
        """
        if not self.is_valid_face(face):
            return False
        
        else:
            return len(np.unique(face)) == 1


    def get_num_solved_faces(self) -> int:
        """Counts the number of solved faces by examining each one using
        :func:`is_solved_face` if the currently initialized :class:`Cube`
        :func:`is_well_formed`, 0 otherwise.

        Returns:
            int: **num_faces_solved** - The number of solved faces on the 
            currently initialized :class:`Cube`.

        """

        num_faces_solved = 0

        if self.is_well_formed():

            for face in self.faces:
                if self.is_solved_face(self.faces[faces]):
                    num_faces_solved += 1

        return num_faces_solved


    #==========================================================================
    #       MOVE METHOD(s)
    #==========================================================================
    def move_up(self):
        """Up Move
        """
        if self.is_well_formed():

            self.faces['UP_FACE'] = np.rot90(self.faces['UP_FACE'], 
                                             axes=(1,0))

            temp = self.faces['BACK_FACE'][0,:].copy()

            self.faces['BACK_FACE'][0,:] = self.faces['LEFT_FACE'][0,:]
            self.faces['LEFT_FACE'][0,:] = self.faces['FRONT_FACE'][0,:]
            self.faces['FRONT_FACE'][0,:] = self.faces['RIGHT_FACE'][0,:]

            self.faces['RIGHT_FACE'][0,:] = temp


    def move_up_inverse(self):
        """Up Inverse Move
        """
        if self.is_well_formed():

            self.faces['UP_FACE'] = np.rot90(self.faces['UP_FACE'])

            temp = self.faces['BACK_FACE'][0,:].copy()
            self.faces['BACK_FACE'][0,:] = self.faces['RIGHT_FACE'][0,:]
            self.faces['RIGHT_FACE'][0,:] = self.faces['FRONT_FACE'][0,:]
            self.faces['FRONT_FACE'][0,:] = self.faces['LEFT_FACE'][0,:]
            self.faces['LEFT_FACE'][0,:] = temp


    def move_down(self):
        """Down Move
        """
        if self.is_well_formed():

            self.faces['DOWN_FACE'] = np.rot90(self.faces['DOWN_FACE'])

            temp = self.faces['FRONT_FACE'][2,:].copy()
            self.faces['FRONT_FACE'][2,:] = self.faces['LEFT_FACE'][2,:]
            self.faces['LEFT_FACE'][2,:] = self.faces['BACK_FACE'][2,:]
            self.faces['BACK_FACE'][2,:] = self.faces['RIGHT_FACE'][2,:]
            self.faces['RIGHT_FACE'][2,:] = temp


    def move_down_inverse(self):
        """Down Inverse Move
        """
        if self.is_well_formed():

            self.faces['DOWN_FACE'] = np.rot90(self.faces['DOWN_FACE'], 
                                               axes=(1,0))

            temp = self.faces['FRONT_FACE'][2,:].copy()
            self.faces['FRONT_FACE'][2,:] = self.faces['RIGHT_FACE'][2,:]
            self.faces['RIGHT_FACE'][2,:] = self.faces['BACK_FACE'][2,:]
            self.faces['BACK_FACE'][2,:] = self.faces['LEFT_FACE'][2,:]
            self.faces['LEFT_FACE'][2,:] = temp


    def move_front(self):
        """Front Move
        """
        if self.is_well_formed():

            self.faces['FRONT_FACE'] = np.rot90(self.faces['FRONT_FACE'], 
                                                axes=(1,0))

            temp = self.faces['UP_FACE'][2,:].copy()
            self.faces['UP_FACE'][2,:] = np.flip(self.faces['LEFT_FACE'][:,2])
            self.faces['LEFT_FACE'][:,2] = self.faces['DOWN_FACE'][0,:]
            self.faces['DOWN_FACE'][0,:] = np.flip(self.faces['RIGHT_FACE'][:,0])
            self.faces['RIGHT_FACE'][:,0] = temp


    def move_front_inverse(self):
        """Front Inverse Move
        """
        if self.is_well_formed():

            self.faces['FRONT_FACE'] = np.rot90(self.faces['FRONT_FACE'])

            temp = self.faces['UP_FACE'][2,:].copy()
            self.faces['UP_FACE'][2,:] = self.faces['RIGHT_FACE'][:,0]
            self.faces['RIGHT_FACE'][:,0] = np.flip(self.faces['DOWN_FACE'][0,:])
            self.faces['DOWN_FACE'][0,:] = self.faces['LEFT_FACE'][:,2]
            self.faces['LEFT_FACE'][:,2] = np.flip(temp)


    def move_back(self):
        """Back Move
        """
        if self.is_well_formed():

            self.faces['BACK_FACE'] = np.rot90(self.faces['BACK_FACE'])

            temp = self.faces['DOWN_FACE'][2,:].copy()
            self.faces['DOWN_FACE'][2,:] = self.faces['LEFT_FACE'][:,0]
            self.faces['LEFT_FACE'][:,0] = np.flip(self.faces['UP_FACE'][0,:])
            self.faces['UP_FACE'][0,:] = self.faces['RIGHT_FACE'][:,2]
            self.faces['RIGHT_FACE'][:,2] = np.flip(temp)


    def move_back_inverse(self):
        """Back Inverse Move
        """
        if self.is_well_formed():

            self.faces['BACK_FACE'] = np.rot90(self.faces['BACK_FACE'], 
                                               axes=(1,0))

            temp = self.faces['DOWN_FACE'][2,:].copy()
            self.faces['DOWN_FACE'][2,:] = self.faces['RIGHT_FACE'][:,2]
            self.faces['RIGHT_FACE'][:,2] = self.faces['UP_FACE'][0,:]
            self.faces['UP_FACE'][0,:] = self.faces['LEFT_FACE'][:,0]
            self.faces['LEFT_FACE'][:,0] = temp


    def move_left(self):
        """Left Move
        """
        if self.is_well_formed():

            self.faces['LEFT_FACE'] = np.rot90(self.faces['LEFT_FACE'], 
                                               axes=(1,0))
            
            temp = self.faces['DOWN_FACE'][:,0].copy()
            self.faces['DOWN_FACE'][:,0] = self.faces['FRONT_FACE'][:,0]
            self.faces['FRONT_FACE'][:,0] = self.faces['UP_FACE'][:,0]
            self.faces['UP_FACE'][:,0] = np.flip(self.faces['BACK_FACE'][:,2])
            self.faces['BACK_FACE'][:,2] = np.flip(temp)


    def move_left_inverse(self):
        """Left Inverse Move
        """
        if self.is_well_formed():

            self.faces['LEFT_FACE'] = np.rot90(self.faces['LEFT_FACE'])

            temp = self.faces['DOWN_FACE'][:,0].copy()
            self.faces['DOWN_FACE'][:,0] = np.flip(self.faces['BACK_FACE'][:,2])
            self.faces['BACK_FACE'][:,2] = np.flip(self.faces['UP_FACE'][:,0])
            self.faces['UP_FACE'][:,0] = self.faces['FRONT_FACE'][:,0] 
            self.faces['FRONT_FACE'][:,0] = temp


    def move_right(self):
        """Right Move
        """
        if self.is_well_formed():

            self.faces['RIGHT_FACE'] = np.rot90(self.faces['RIGHT_FACE'], 
                                                axes=(1,0))

            temp = self.faces['UP_FACE'][:,2].copy()
            self.faces['UP_FACE'][:,2] = self.faces['FRONT_FACE'][:,2]
            self.faces['FRONT_FACE'][:,2] = self.faces['DOWN_FACE'][:,2]
            self.faces['DOWN_FACE'][:,2] = np.flip(self.faces['BACK_FACE'][:,0])
            self.faces['BACK_FACE'][:,0] = np.flip(temp)


    def move_right_inverse(self):
        """Right Inverse Move
        """
        if self.is_well_formed():

            self.faces['RIGHT_FACE'] = np.rot90(self.faces['RIGHT_FACE'])

            temp = self.faces['UP_FACE'][:,2].copy()
            self.faces['UP_FACE'][:,2] = np.flip(self.faces['BACK_FACE'][:,0])
            self.faces['BACK_FACE'][:,0] = np.flip(self.faces['DOWN_FACE'][:,2])
            self.faces['DOWN_FACE'][:,2] = self.faces['FRONT_FACE'][:,2]
            self.faces['FRONT_FACE'][:,2] = temp


    def move_middle(self):
        """Middle Slice Move
        """
        if self.is_well_formed():

            temp = self.faces['DOWN_FACE'][:,1].copy()
            self.faces['DOWN_FACE'][:,1] = self.faces['FRONT_FACE'][:,1]
            self.faces['FRONT_FACE'][:,1] = self.faces['UP_FACE'][:,1]
            self.faces['UP_FACE'][:,1] = np.flip(self.faces['BACK_FACE'][:,1])
            self.faces['BACK_FACE'][:,1] = np.flip(temp)


    def move_middle_inverse(self):
        """Middle Slice Inverse Move
        """
        if self.is_well_formed():

            temp = self.faces['DOWN_FACE'][:,1].copy()
            self.faces['DOWN_FACE'][:,1] = np.flip(self.faces['BACK_FACE'][:,1])
            self.faces['BACK_FACE'][:,1] = np.flip(self.faces['UP_FACE'][:,1])
            self.faces['UP_FACE'][:,1] = self.faces['FRONT_FACE'][:,1] 
            self.faces['FRONT_FACE'][:,1] = temp


    def move_equator(self):
        """Equator Slice Move
        """
        if self.is_well_formed():

            temp = self.faces['FRONT_FACE'][1,:].copy()
            self.faces['FRONT_FACE'][1,:] = self.faces['LEFT_FACE'][1,:]
            self.faces['LEFT_FACE'][1,:] = self.faces['BACK_FACE'][1,:]
            self.faces['BACK_FACE'][1,:] = self.faces['RIGHT_FACE'][1,:]
            self.faces['RIGHT_FACE'][1,:] = temp


    def move_equator_inverse(self):
        """Equator Slice Inverse Move
        """
        if self.is_well_formed():

            temp = self.faces['FRONT_FACE'][1,:].copy()
            self.faces['FRONT_FACE'][1,:] = self.faces['RIGHT_FACE'][1,:]
            self.faces['RIGHT_FACE'][1,:] = self.faces['BACK_FACE'][1,:]
            self.faces['BACK_FACE'][1,:] = self.faces['LEFT_FACE'][1,:]
            self.faces['LEFT_FACE'][1,:] = temp


    def move_standing(self):
        """Standing Slice Move
        """
        if self.is_well_formed():

            temp = self.faces['UP_FACE'][1,:].copy()
            self.faces['UP_FACE'][1,:] = np.flip(self.faces['LEFT_FACE'][:,1])
            self.faces['LEFT_FACE'][:,1] = self.faces['DOWN_FACE'][1,:]
            self.faces['DOWN_FACE'][1,:] = np.flip(self.faces['RIGHT_FACE'][:,1])
            self.faces['RIGHT_FACE'][:,1] = temp


    def move_standing_inverse(self):
        """Standing Slice Inverse Move
        """
        if self.is_well_formed():

            temp = self.faces['UP_FACE'][1,:].copy()
            self.faces['UP_FACE'][1,:] = self.faces['RIGHT_FACE'][:,1]
            self.faces['RIGHT_FACE'][:,1] = np.flip(self.faces['DOWN_FACE'][1,:])
            self.faces['DOWN_FACE'][1,:] = self.faces['LEFT_FACE'][:,1]
            self.faces['LEFT_FACE'][:,1] = np.flip(temp)


    def rotate_pitch(self):
        """Pitch Rotation
        """
        if self.is_well_formed():

            self.move_left_inverse()
            self.move_middle_inverse()
            self.move_right()


    def rotate_pitch_inverse(self):
        """Pitch Inverse Rotation
        """
        if self.is_well_formed():

            self.move_left()
            self.move_middle()
            self.move_right_inverse()


    def rotate_roll(self):
        """Roll Rotation
        """
        if self.is_well_formed():

            self.move_front()
            self.move_standing()
            self.move_back_inverse()


    def rotate_roll_inverse(self):
        """Roll Inverse Rotation
        """
        if self.is_well_formed():

            self.move_front_inverse()
            self.move_standing_inverse()
            self.move_back()


    def rotate_yaw(self):
        """Yaw Rotation
        """
        if self.is_well_formed():

            self.move_up()
            self.move_equator_inverse()
            self.move_down_inverse()


    def rotate_yaw_inverse(self):
        """Yaw Inverse Rotation
        """
        if self.is_well_formed():

            self.move_up_inverse()
            self.move_equator()
            self.move_down()

    #==========================================================================
    #       CONSTANTS FOR DEFAULT CUBE-COLORS
    #==========================================================================
    DEFAULT_UP_COLOR = '#ffffff'        # White
    DEFAULT_DOWN_COLOR = '#ffd500'      # Cyber Yellow
    DEFAULT_FRONT_COLOR = '#009b48'     # Green (Pigment)
    DEFAULT_BACK_COLOR = '#0045ad'      # Cobalt Blue
    DEFAULT_LEFT_COLOR = '#ff5900'      # Orange (Pantone)
    DEFAULT_RIGHT_COLOR = '#b90000'     # UE Red

    DEFAULT_FACE_COLORS = {
        'UP_COLOR' : DEFAULT_UP_COLOR,
        'DOWN_COLOR' : DEFAULT_DOWN_COLOR,
        'FRONT_COLOR' : DEFAULT_FRONT_COLOR,
        'BACK_COLOR' : DEFAULT_BACK_COLOR,
        'LEFT_COLOR' : DEFAULT_LEFT_COLOR,
        'RIGHT_COLOR' : DEFAULT_RIGHT_COLOR
    }

    DEFAULT_UP_FACE = np.full((3,3), DEFAULT_UP_COLOR)
    DEFAULT_DOWN_FACE = np.full((3,3), DEFAULT_DOWN_COLOR)
    DEFAULT_FRONT_FACE = np.full((3,3), DEFAULT_FRONT_COLOR)
    DEFAULT_BACK_FACE = np.full((3,3), DEFAULT_BACK_COLOR)
    DEFAULT_LEFT_FACE = np.full((3,3), DEFAULT_LEFT_COLOR)
    DEFAULT_RIGHT_FACE = np.full((3,3), DEFAULT_RIGHT_COLOR)

    DEFAULT_FACES = {
        'UP_FACE' : DEFAULT_UP_FACE,
        'DOWN_FACE' : DEFAULT_DOWN_FACE,
        'FRONT_FACE' : DEFAULT_FRONT_FACE,
        'BACK_FACE' : DEFAULT_BACK_FACE,
        'LEFT_FACE' : DEFAULT_LEFT_FACE,
        'RIGHT_FACE' : DEFAULT_RIGHT_FACE
    }

